local L = AceLibrary("AceLocale-2.2"):new("CastYeller")
local BS = AceLibrary("Babble-Spell-2.2")
local AceOO = AceLibrary("AceOO-2.0")

L:RegisterTranslations("enUS", function() return {
	['Spells'] = true,
	['Enable announces for selected spells.'] = true,
	['Channels'] = true,
	['Enable announces to the selected channels.'] = true,
	['Only in instance'] = true,
	['Only announce spell cast in instance.'] = true,
	['Enable %s announce.'] = true,
	['Not in battleground'] = true,
	['Disable announces in battleground.'] = true,
	['Funny announces'] = true,
	['Use funny announces when available.'] = true,
	['Custom Channel'] = true,
	['Send announces to the following custom channel.'] = true,
	['<channelName>'] = true,
	['Please define a custom channel or disable custom channel announces.'] = true,
	['you'] = true,

	['$SPELL on $TARGET'] = '$SPELL on $TARGET',
	['$SPELL'] = '$SPELL',
	['GROUP-chat'] = 'Raid or party',
	['BATTLEGROUND-chat'] = 'Battleground',
	['CHANNEL-chat'] = 'Custom Channel',
	['BROADCAST-chat'] = 'Broadcast',
} end)

L:RegisterTranslations("frFR", function() return {
	['Spells'] = 'Sorts',
	['Enable announces for selected spells.'] = 'Active les annones pour les sorts sélectionnés.',
	['Channels'] = 'Canaux',
	['Enable announces to the selected channels.'] = 'Active les announces sur les canaux sélectionnés.',
	['Only in instance'] = 'En instance uniquement',
	['Only announce spell cast in instance.'] = "N'annonce les sorts qu'en instance.",
	['Enable %s announce.'] = 'Activer les announces sur %s.',
	['Not in battleground'] = 'Pas en champ de bataille',
	['Disable announces in battleground.'] = 'Désactive les announces sur les champs de bataille.',
	['Funny announces'] = 'Annonces rigolotes',
	['Use funny announces when available.'] = 'Utilise des annonces rigolotes quand il y en a de disponibles.',
	['you'] = 'toi',
	
	['$SPELL on $TARGET'] = '$SPELL sur $TARGET',
	['GROUP-chat'] = 'Raid ou groupe',
	['BATTLEGROUND-chat'] = 'Champs de bataille',
	['CHANNEL-chat'] = 'Canal personnalisé',
	['BROADCAST-chat'] = 'Diffusion',
} end)

CastYeller = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0", "AceEvent-2.0", "AceDB-2.0"
--[[, "Parser-3.0"]]
)
local CastYeller = CastYeller
if AceLibrary:HasInstance('AceDebug-2.0') then
	AceLibrary('AceDebug-2.0'):embed(CastYeller)
	CastYeller:SetDebugging(false)
else
	function CastYeller.Debug() end
end
CastYeller.revision = tonumber(string.sub("$Revision: 29774 $", 12, -3))
CastYeller.version = "r" .. CastYeller.revision
CastYeller.date = string.sub("$Date: 2007-03-08 16:48:27 -0500 (Thu, 08 Mar 2007) $", 8, 17)

local function erase(...)
	for i=1,select('#', ...) do
		local t = select(i, ...)
		for k in pairs(t) do t[k] = nil end
	end
end

-- Privates
local classesAnnounces = {}
local announces = {}
local knownSpells = {}
local usedChannels = {}
local hasFunnyAnnounces = false
local yellCooldown = {}
local events = {
	start = {},
	success = {},
--	hit = {},
--	miss = {},
}

-- Functions
local validateAnnounce, setupChannelOptions, setupSpellOptions

-- Addon setup
CastYeller:RegisterDB("CastYellerDB")

CastYeller.options = {	
	type = 'group',
	args = {
		spell = {
			name = L['Spells'], 
			desc = L['Enable announces for selected spells.'], 
			type = 'group',				
			args = {},
			order = 100,
			disabled = '~IsActive',
		},
		channel = {
			name = L['Channels'],
			desc = L['Enable announces to the selected channels.'], 
			type = 'group',
			args = {},
			order = 110,
			disabled = '~IsActive',
		},
		customChannel = {
			name = L['Custom Channel'],
			desc = L['Send announces to the following custom channel.'],
			type = 'text',
			get = function() return CastYeller.db.profile.customChannel end,
			set = function(v) CastYeller.db.profile.customChannel = v end,
			validate = function(v) return v:match('^%w+$') end,
			usage = L['<channelName>'],
			order = 120,
			disabled = function() return not CastYeller:IsActive() or not CastYeller.db.profile.channels.CHANNEL end,
			hidden = function() return not usedChannels.CHANNEL end,
		},
		funny = {
			name = L['Funny announces'],
			desc = L['Use funny announces when available.'],
			type = 'toggle',
			get = function() return hasFunnyAnnounces and CastYeller.db.profile.funnyAnnounces end,
			set = function(v) CastYeller.db.profile.funnyAnnounces = v end,
			order = 200,
			disabled = '~IsActive',
			hidden = function() return not hasFunnyAnnounces end,
		},
		instanceOnly = {
			name = L['Only in instance'],
			desc = L['Only announce spell cast in instance.'],
			type = 'toggle',
			get = function() return CastYeller.db.profile.instanceOnly end,
			set = function(v) CastYeller.db.profile.instanceOnly = v end,
			order = 210,
			disabled = '~IsActive',
		},
		noBattleground = {
			name = L['Not in battleground'],
			desc = L['Disable announces in battleground.'],
			type = 'toggle',
			get = function() return CastYeller.db.profile.noBattleground end,
			set = function(v) CastYeller.db.profile.noBattleground = v end,
			order = 220,
			disabled = '~IsActive',
		},
	}
}

-- private constants
local defaultAnnounces
do
		local basicAnnounces = {
			targetted = L['$SPELL on $TARGET'],
			untargetted = L['$SPELL'],
		}
		defaultAnnounces = {
			start = basicAnnounces,
			success = basicAnnounces,
--			hit = basicAnnounces,
--			miss = basicAnnounces,
		}
end

local conditions = {
	combat       = function() return UnitAffectingCombat('player') end,
	party        = function() return (GetNumPartyMembers() or 0) > 0 end,
	raid         = function() return (GetNumRaidMembers() or 0) > 0 end,
	battleground = function() return (GetNumBattlefieldScores() or 0) > 0 end,
	instance     = function()
		local inInstance, iType = IsInInstance()
		return inInstance and (iType == 'party' or iType == 'raid')
	end,
}

local undefinedCustomAlert

local channels = {
  EMOTE        = true,
	BATTLEGROUND = true,
	PARTY        = true,
	RAID         = true,
	CHANNEL      = true,
	WHISPER      = true,
	
	_BROADCAST    = { '_CHANNEL', '_GROUP' },
	
	_CHANNEL      = function(target)
		local success, id = pcall(GetChannelName, CastYeller.db.profile.customChannel)
		if success and id and id > 0 then
			return 'CHANNEL', id
		elseif not undefinedCustomAlert then
			-- Warn one time
			CastYeller:Print(L['Please define a custom channel or disable custom channel announces.'])
			undefinedCustomAlert = true
		end
	end,
	
	_WHISPER     = function(target)
		local success, result = pcall(UnitExists, target)
		if success and result and UnitIsPlayer(target) then
			return 'WHISPER', target
		else
			CastYeller:Debug('Cannot whisper to %q', target)
		end
	end,
	
	_GROUP        = function() 
		if (GetNumBattlefieldScores() or 0) > 0 then
			return 'BATTLEGROUND'
		elseif (GetNumRaidMembers() or 0) > 0 then
			return 'RAID'
		elseif (GetNumPartyMembers() or 0) > 0 then
			return 'PARTY'
		else
			CastYeller:Debug('GROUP announces are ignored in solo.')
		end
	end,
}

-- Translate channel key to effective list of potential channels
local flatChannels = {
  EMOTE        = true,
	BATTLEGROUND = true,
	PARTY        = true,
	RAID         = true,
	CHANNEL      = true,
	WHISPER      = true,
	_CHANNEL      = { 'CHANNEL' } ,
	_WHISPER      = { 'WHISPER' } ,
	_BROADCAST    = { 'CHANNEL', 'BATTLEGROUND', 'PARTY', 'RAID' },
	_GROUP        = { 'BATTLEGROUND', 'PARTY', 'RAID' },
}

function CastYeller:RegisterAnnounces(class, func)
	classesAnnounces[class] = func
end

function CastYeller:OnInitialize()

	self:RegisterDefaults('profile', {
		instanceOnly = false,
		noBattleground = true,
		funnyAnnounces = true,
		channels = {
			['*'] = true,
		},
	})
	self:RegisterDefaults('class', {spells = {}})
	
	self:RegisterChatCommand({"/cy", "/castyeller"}, self.options)

	local classFunc = classesAnnounces[select(2, UnitClass('player'))]
	if not classFunc then
		return
	end
	for spell, announce in pairs(classFunc()) do
		if validateAnnounce(spell, announce) then
			announces[spell] = announce
		end
	end
	erase(classesAnnounces)
	
	setupChannelOptions(self.options.args.channel.args)
	setupSpellOptions(self.options.args.spell.args)
end

function CastYeller:OnEnable()
	if announces and next(announces) then
		self:RegisterEvent('SPELLS_CHANGED', 'SetupEvents', 1)
		self:RegisterEvent('CastYeller_Announce', 'Yell')
		self:SetupEvents()
	end
end

-- Spell recognition

local function englishSpell(spell)
	return BS:HasReverseTranslation(spell) and BS:GetReverseTranslation(spell) or spell
end

local function ScanSpellbook()
	erase(knownSpells)
	local i = 1
	while true do
		local spell = GetSpellName(i, BOOKTYPE_SPELL)
		if not spell then
			break
		end
		knownSpells[englishSpell(spell)] = spell
		i = i + 1
	end
end

-- Setup event listening

local function ToggleEvent(self, condition, eventName, ...)
	local registered = self:IsEventRegistered(eventName)
	if condition and not registered then
		self:Debug('Registering event '..eventName)
		self:RegisterEvent(eventName, ...)
	elseif not condition and registered then
		self:Debug('Unregistering event '..eventName)
		self:UnregisterEvent(eventName)
	end
end

local function usesChannel(channel)
	if type(channel) == 'table' then
		for i,channel in ipairs(channel) do
			usesChannel(channel)
		end
	elseif flatChannels[channel] ~= true then
		return usesChannel(flatChannels[channel])
	end
	
	usedChannels[channel] = true
end

function CastYeller:SetupEvents()	
	erase(usedChannels)
	for event in pairs(events) do
		erase(events[event])
	end
	ScanSpellbook()
	hasFunnyAnnounces = false
		
	for spell, announce in pairs(announces) do
		local locSpell = knownSpells[spell]
		if locSpell then
			local used = false
			local spellKey = announce.spellKey or spell
			for event,values in pairs(announce.events) do
				self:Debug('announce on %s: %s', event, spell)
				events[event][locSpell] = spellKey
				if values ~= true then
					hasFunnyAnnounces = true
				end
			end
			usesChannel(announce.channel)
		else
			self:Debug('unknown spell: '..spell)
		end
	end
	
	ToggleEvent(self, next(events.start), 'SpellEvents_Player_Spell_Start', 'OnSpellStart')
	ToggleEvent(self, next(events.start), 'SpellEvents_Player_Channel_Start', 'OnSpellStart')
	ToggleEvent(self, next(events.success), 'SpellEvents_Player_Instant_Success', 'OnInstantSpellSuccess')
	ToggleEvent(self, next(events.success), 'SpellEvents_Player_Spell_Success', 'OnSpellSuccess')
	
	--[[
	self:UnregisterAllParserEvents()
	if next(events.hit) then
		self:RegisterParserEvent(
			{ eventType = 'Damage', sourceID = "player", abilityName_in = events.hit }, 
			'ParserEvent'
		)
	end
	if next(events.miss) then
		self:RegisterParserEvent(
			{ eventType = 'Miss', sourceID = "player", abilityName_in = events.miss }, 
			'ParserEvent'
		)
	end
	--]]
end

-- Announce stuff

local function checkConditions(self, condTable, spell, target)
	for i,cond in pairs(condTable) do
		local reverse = false
		if cond:match('^no(.+)$') then
			reverse, cond = true, cond:match('^no(.+)$')
		end
		local result = conditions[cond](spell, target) and true or false
		if result == reverse then
			self:Debug('Condition %s failed (%q, should be %q)', cond, result, not reverse)
			return false
		end
	end
	return true
end

local function additionnalTags()
	local pClass = select(2,UnitClass('player'))
	if pClass == 'HUNTER' or pClass == 'MAGE' then
		if UnitExists('pet') then
			return 'PET', UnitName('pet')
		end
	end
end

local function validateArg(text)
	return type(text) == 'string' and text:len() > 0 and text
end

local function formatAnnounce(text, spell, target, ...)
	target = validateArg(target)
	if target then
		text = text:gsub('%$TARGET', target)
	end
	if spell then
		text = text:gsub('%$SPELL', spell)
	end
	if select('#', ...) > 1 then
		for i = 1,select('#', ...),2 do
			local k,v = select(i, ...)
			v = validateArg(v)
			if k and v then
				text = text:gsub('%$'..k, v)
			end
		end
	end
	return text
end

local function getDefaultAnnounce(event, target)
	return validateArg(target) and defaultAnnounces[event].targetted or defaultAnnounces[event].untargetted
end

local function sendAnnounce(self, channel, announce, target, spell, spellKey, event, sendTarget) 
	if not channel then 
		return self:Debug('No channel to announce to (%q-%q)', spell, event)
	elseif type(channel) == 'table' then
		self:Debug('multi-announce for %q-%q', spell, event)
		for i,subChannel in pairs(channel) do
			sendAnnounce(self, subChannel, announce, target, spell, spellKey, event, sendTarget) 
		end
		return
	elseif not self.db.profile.channels[channel] then		
		return self:Debug('Announces to %q are disabled', channel)
	elseif type(channels[channel]) == 'function' then
		local newChannel, newSendTarget = channels[channel](target, channel, sendTarget)
		self:Debug('dynamic announce for %q-%q => channel: %q, send target: %q', spell, event, newChannel, newSendTarget)
		return sendAnnounce(self, newChannel, announce, target, spell, spellKey, event, newSendTarget or sendTarget) 
	elseif type(channels[channel]) == 'table' then
		self:Debug('broadcast announce for %q-%q ', spell, event)
		for i,subChannel in pairs(channels[channel]) do
			sendAnnounce(self, subChannel, announce, target, spell, spellKey, event, sendTarget) 
		end
		return
	elseif not channel then
		return self:Debug('No channel to announce to')
	end
	local text
	if channel == 'WHISPER' then
		if sendTarget then
			text = '<<< '..formatAnnounce(getDefaultAnnounce(event, target), spell, L['you'], additionnalTags())..' >>>'
		else
			return self:Debug('No target to whisper to')
		end
	else
		text = announce.events[event]
		if text == true or not CastYeller.db.profile.funnyAnnounces then
			text = getDefaultAnnounce(event, target)
		elseif type(text) == 'table' then
			text = text[math.random(1,#text)]
		end
		text = formatAnnounce(text, spell, target, additionnalTags())
		if channel ~= 'EMOTE' then
			text = '<<< '..text..' >>>'
		end
	end
	self:Debug('Sending %q announce to %q (%q)', spellKey, channel, text)
	SendChatMessage(text, channel, nil, sendTarget)
	yellCooldown[spellKey] = GetTime() + 5
end

function CastYeller:Yell(spell, target, event)
	local spellKey = events[event] and events[event][spell]
	if not spellKey then
		return
	end
	local announce = announces[spellKey]
	self:Debug('%q(%q)', event, spellKey)
	if self.db.profile.instanceOnly and not conditions.instance() then
		return self:Debug('Announces only in instances')
	elseif self.db.profile.noBattleground and conditions.battleground() then
		return self:Debug('No announces only in battleground')
	elseif not self:IsAnnouncing(spellKey) then
		return self:Debug('%q announces are disabled', spellKey)
	elseif announce.conditions and not checkConditions(self, announce.conditions, spell, target) then
		return 
	elseif yellCooldown[spellKey] and yellCooldown[spellKey] > GetTime() then
		return self:Debug('%q announce is on cooldown', spellKey)
	else
		return sendAnnounce(self, announce.channel, announce, target, spell, spellKey, event)
	end
end

-- Spell event handlers

function CastYeller:OnSpellStart(spell, rank, name, icon, startTime, endTime, isTradeSkill, target)
	return self:TriggerEvent('CastYeller_Announce', spell, target, 'start')
end

function CastYeller:OnInstantSpellSuccess(spell, rank, target)
	return self:TriggerEvent('CastYeller_Announce', spell, target, 'success')
end

function CastYeller:OnSpellSuccess(spell, rank, name, icon, startTime, endTime, isTradeSkill, target)
	return self:TriggerEvent('CastYeller_Announce', spell, target, 'success')
end

--[[
function CastYeller:ParserEvent(e)
	return self:TriggerEvent('CastYeller_Announce', e.abilityName, e.targetName, (e.eventType == 'Damage') and 'hit' or 'miss')
end
--]]

-- Options handling

function CastYeller:IsAnnouncing(spell)
	local s = self.db.class.spells
	if s[spell] == nil then
		s[spell] = not announces[spell].defaultDisabled
	end
	return s[spell]
end

function CastYeller:ToggleAnnounce(spell, value)
	if value == nil then
		value = not self:IsAnnouncing(spell)
	end
	self.db.class.spells[spell] = value
end

function setupSpellOptions(options)
	-- Setup options too
	erase(options)
	for spell, announce in pairs(announces) do
		local locSpell = announce.spellDisplay or BS[spell]
		local spell = spell
		local spellKey = announce.spellKey or spell
		options[spellKey] = {
			name = locSpell,
			desc = string.format('Enable announces for %s.', locSpell),
			type = 'toggle',
			get = function() return knownSpells[spell] and CastYeller:IsAnnouncing(spellKey) end,
			set = function(v) return CastYeller:ToggleAnnounce(spellKey, v) end,
			disabled = function() return not knownSpells[spell] end,
		}
	end
end

function setupChannelOptions(options)
	erase(options)
	for k in pairs(channels) do
		if not k:match('^_') then
			local k = k		
			local label = getglobal(k..'_MESSAGE') or L[k..'-chat']
			options[k] = {
				name = label,
				desc = L['Enable %s announce.']:format(label:lower()),
				type = 'toggle',
				get = function() return CastYeller.db.profile.channels[k] end,
				set = function(v) CastYeller.db.profile.channels[k] = v end,
				hidden = function() return not usedChannels[k] end,
			}
		end
	end
end

-- Validate announce data

function validateAnnounce(spell, announce) -- it's already local
	local _error = error
	local function error(msg)
		_error(spell..': '..msg)
		return false
	end
	
	if not BS:HasTranslation(spell) then
		return error('no translation')
	end
	
	if type(announce.events) ~= 'table' or not next(announce.events) then
		return error('.events is mandatory and should be a table')
	end
	for event,value in pairs(announce.events) do
		if type(event) ~= 'string' or not events[event] then
			return error(string.format('invalid event: %q', tostring(event)))
		elseif type(value) ~= 'string' and type(value) ~= 'table' and value ~= true then
			return error('invalid event data, should be a string, a table of strings or true')
		end
	end
		
	if announce.conditions then
		if type(announce.conditions) == 'string' then
			announce.conditions = { announce.conditions }
		elseif type(announce.conditions) ~= 'table' then
			return error('.conditions should be a table of strings')
		end
		for i,c in ipairs(announce.conditions) do
			if type(c) ~= 'string' then
				return error('.conditions should be a table of strings')
			end
			if c:match('^no(.*)$') then
				c = c:match('^no(.*)$')
			end
			if not conditions[c] then
				return error('condition '..c..' does not exist')
			end			
		end
	end

	if announce.channel then
		if type(announce.channel) == 'table' then
			for i,c in ipairs(announce.channel) do
				if type(c) ~= 'string' then
					return error('.channel should be a string or table of strings')
				end
				if not channels[c] then
					return error('channel '..c..' does not exist')
				end
			end
		elseif type(announce.channel) ~= 'string' then
			return error('channel should be a string or a table of strings')
		end
	else
		return error('.channel is mandatory')
	end
	
	announce.spell = spell

	return true
end
