--[[
Name: SpellEvents-1.0
Revision: $Rev: 28936 $
Author: Kameril
Website: http://www.wowace.com/wiki/SpellEvents-1.0
Documentation: http://www.wowace.com/wiki/SpellEvents-1.0
SVN: svn://svn.wowace.com/root/trunk/SpellEventsLib/
Description: A library that aims to make sense of Blizzard's new spell events.
Dependencies: AceLibrary, AceEvent-2.0
]]


local MAJOR_VERSION = "SpellEvents-1.0"
local MINOR_VERSION = "$Revision: 28936 $"

if not AceLibrary then error(MAJOR_VERSION .. " requires AceLibrary") end
if not AceLibrary:IsNewVersion(MAJOR_VERSION, MINOR_VERSION) then return end

local Spell = {}
local UnitCastingInfo, UnitChannelInfo = UnitCastingInfo, UnitChannelInfo

local pSpell, pRank, pDisplayName, pIcon, pStartTime, pEndTime, pIsTradeSkill --variables used to save the player's current casts
function Spell:UNIT_SPELLCAST_START(unit)
	local spell, rank, displayName, icon, startTime, endTime, isTradeSkill = UnitCastingInfo(unit)
	if not spell then return end
	if unit == "player" then
		self.PlayerInstants[spell] = nil
		self.PlayerChannels[spell] = nil
		
		pSpell, pRank, pDisplayName, pIcon, pStartTime, pEndTime, pIsTradeSkill = spell, rank, displayName, icon, startTime, endTime, isTradeSkill
		self:TriggerEvent("SpellEvents_Player_Spell_Start", pSpell, pRank, pDisplayName, pIcon, pStartTime, pEndTime, pIsTradeSkill, self.PlayerCasts[spell])
	else
		self.UnitSpell[unit] = spell
		self.UnitChannel[unit] = nil
		self.UnitRank[unit] = rank
		self.UnitDisplay[unit] = displayName
		self.UnitIcon[unit] = icon
		self.UnitStartTime[unit] = startTime
		self.UnitEndTime[unit] = endTime
		self.UnitIsTradeSkill[unit] = isTradeSkill
		self:TriggerEvent("SpellEvents_Unit_Spell_Start", unit, spell, rank, displayName, icon, startTime, endTime, isTradeSkill)
	end
end


function Spell:UNIT_SPELLCAST_STOP(unit)
	if unit == "player" then
		if pSpell and not self.PlayerChannels[pSpell] then
			self:TriggerEvent("SpellEvents_Player_Spell_Interrupted", pSpell, pRank, pDisplayName, pIcon, pStartTime, pEndTime, pIsTradeSkill, self.PlayerCasts[pSpell])
			self.PlayerChannels[pSpell] = nil
			pSpell, pRank, pDisplayName, pIcon, pStartTime, pEndTime, pIsTradeSkill = nil
		end
	elseif self.UnitSpell[unit] then
		self:TriggerEvent("SpellEvents_Unit_Spell_Stop", unit, self.UnitSpell[unit], self.UnitRank[unit], 
			self.UnitDisplay[unit], self.UnitIcon[unit], self.UnitStartTime[unit], self.UnitEndTime[unit], self.UnitIsTradeSkill[unit])
		self.UnitSpell[unit], self.UnitRank[unit], self.UnitDisplay[unit], self.UnitIcon[unit], 
		self.UnitStartTime[unit], self.UnitEndTime[unit], self.UnitIsTradeSkill[unit] = nil
	end
end

do
	local spell, reason
	function Spell:CHAT_MSG_SPELL_FAILED_LOCALPLAYER(msg)
		spell, reason = msg:match("^You fail to %w- (.+): (.+).") --convert this to deformat/localization
	end
	function Spell:UNIT_SPELLCAST_FAILED(unit) --This event isn't received for any unit that isn't the player
		if unit == "player" then
			self:TriggerEvent("SpellEvents_Player_Spell_Failed", spell, reason)
		end
	end
end

--This event isn't received for any unit that isn't the player
function Spell:UNIT_SPELLCAST_INTERRUPTED(unit)
	if unit == "player" then
		if pSpell then
			self:TriggerEvent("SpellEvents_Player_Spell_Interrupted", pSpell, pRank, pDisplayName, pIcon, pStartTime, pEndTime, pIsTradeSkill, self.PlayerCasts[pSpell])
			self.PlayerCasts[pSpell] = nil
			pSpell, pRank, pDisplayName, pIcon, pStartTime, pEndTime, pIsTradeSkill = nil
		end
	end
end

function Spell:UNIT_SPELLCAST_DELAYED(unit)
	local spell,_,_,_, startTime, endTime = UnitCastingInfo(unit)
	if not spell then return end
	if unit == "player" then
		self:TriggerEvent("SpellEvents_Player_Spell_Changed", pSpell, endTime - pEndTime, startTime, endTime)
		pStartTime, pEndTime = startTime, endTime
	elseif self.UnitSpell[unit] then
		self:TriggerEvent("SpellEvents_Unit_Spell_Changed", unit, spell, endTime - self.UnitEndTime[unit], startTime, endTime)
		self.UnitStartTime[unit] = startTime
		self.UnitEndTime[unit] = endTime
	end
end

function Spell:UNIT_SPELLCAST_CHANNEL_START(unit)
	local spell, rank, displayName, icon, startTime, endTime, isTradeSkill = UnitChannelInfo(unit)
	if not spell then return end
	if unit == "player" then
		self.PlayerInstants[spell] = nil
		self.PlayerCasts[spell] = nil
		self.PlayerChannels[spell] = self.PlayerChannels[spell] or ""
		pSpell, pRank, pDisplayName, pIcon, pStartTime, pEndTime, pIsTradeSkill = spell, rank, displayName, icon, startTime, endTime, isTradeSkill
		self:TriggerEvent("SpellEvents_Player_Channel_Start", pSpell, pRank, pDisplayName, pIcon, pStartTime, pEndTime, pIsTradeSkill, self.PlayerCasts[spell])
	else
		self.UnitSpell[unit] = nil
		self.UnitChannel[unit] = spell
		self.UnitRank[unit] = rank
		self.UnitDisplay[unit] = displayName
		self.UnitIcon[unit] = icon
		self.UnitStartTime[unit] = startTime
		self.UnitEndTime[unit] = endTime
		self.UnitIsTradeSkill[unit] = isTradeSkill
		self:TriggerEvent("SpellEvents_Unit_Channel_Start", unit, spell, rank, displayName, icon, startTime, endTime, isTradeSkill)
	end
end

function Spell:UNIT_SPELLCAST_CHANNEL_UPDATE(unit)
	local spell, rank, displayName, icon, startTime, endTime, isTradeSkill = UnitChannelInfo(unit)
	if not spell then return end
	if unit == "player" then
		if pSpell then
			self:TriggerEvent("SpellEvents_Player_Channel_Changed",  pSpell, endTime - pEndTime)
			pStartTime, pEndTime = startTime, endTime
		end
	elseif self.UnitChannel[unit] then
		self:TriggerEvent("SpellEvents_Unit_Channel_Changed", unit, spell, endTime - self.UnitEndTime[unit], startTime, endTime)
		self.UnitStartTime[unit] = startTime
		self.UnitEndTime[unit] = endTime
	end
end

function Spell:UNIT_SPELLCAST_CHANNEL_STOP(unit)
	if unit == "player" then
		if pSpell then
			self:TriggerEvent("SpellEvents_Player_Channel_Stop", pSpell, pRank, pDisplayName, pIcon, pStartTime, pEndTime, pIsTradeSkill, self.PlayerChannels[pSpell])
			self.PlayerChannels[pSpell] = nil
			pSpell, pRank, pDisplayName, pIcon, pStartTime, pEndTime, pIsTradeSkill = nil
		end
	elseif self.UnitChannel[unit] then
		self:TriggerEvent("SpellEvents_Unit_Channel_Stop", unit, self.UnitChannel[unit], self.UnitRank[unit], 
			self.UnitDisplay[unit], self.UnitIcon[unit], self.UnitStartTime[unit], self.UnitEndTime[unit], self.UnitIsTradeSkill[unit])
		self.UnitChannel[unit], self.UnitRank[unit], self.UnitDisplay[unit], self.UnitIcon[unit], 
		self.UnitStartTime[unit], self.UnitEndTime[unit], self.UnitIsTradeSkill[unit] = nil
	end
end

function Spell:UNIT_SPELLCAST_SENT(unit, spell, rank, target)
	pSpell, pRank = spell, rank
	self.PlayerInstants[spell] = target
	self.PlayerCasts[spell] = target
	self.PlayerChannels[spell] = target
end

function Spell:UNIT_SPELLCAST_SUCCEEDED(unit, spell, rank)
	if self.PlayerInstants[spell] then
		self:TriggerEvent("SpellEvents_Player_Instant_Success", spell, rank, self.PlayerInstants[spell])
	elseif self.PlayerCasts[spell] then
		self:TriggerEvent("SpellEvents_Player_Spell_Success", pSpell, pRank, pDisplayName, pIcon, pStartTime, pEndTime, pIsTradeSkill, self.PlayerCasts[spell])
		pSpell, pRank, pDisplayName, pIcon, pStartTime, pEndTime, pIsTradeSkill = nil
	end
	self.PlayerInstants[spell] = nil
	self.PlayerCasts[spell] = nil
end

local function activate(self, oldLib, oldDeactivate)
	self.UnitSpell = {}
	self.UnitRank = {}
	self.UnitDisplay = {}
	self.UnitIcon = {}
	self.UnitStartTime = {}
	self.UnitEndTime = {}
	self.UnitIsTradeSkill = {}
	self.UnitChannel = {}

	self.PlayerCasts = {}
	self.PlayerChannels = {}
	self.PlayerInstants = {}
	
	if oldDeactivate then
		oldDeactivate(oldLib)
	end
end

local function deactivate(self)
	if self.UnregisterAllEvents then
		self:UnregisterAllEvents()
	end
end

local function external(self, major, instance)
	if major == 'AceEvent-2.0' then
		instance:embed(self)
		
		self:RegisterEvent("UNIT_SPELLCAST_START")
		self:RegisterEvent("UNIT_SPELLCAST_STOP")
		self:RegisterEvent("UNIT_SPELLCAST_FAILED")
		self:RegisterEvent("UNIT_SPELLCAST_INTERRUPTED")
		self:RegisterEvent("UNIT_SPELLCAST_DELAYED")
	
		self:RegisterEvent("UNIT_SPELLCAST_CHANNEL_START")
		self:RegisterEvent("UNIT_SPELLCAST_CHANNEL_UPDATE")
		self:RegisterEvent("UNIT_SPELLCAST_CHANNEL_STOP")
	
		self:RegisterEvent("UNIT_SPELLCAST_SENT")
		self:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")
	
		self:RegisterEvent("CHAT_MSG_SPELL_FAILED_LOCALPLAYER")
	end
end

AceLibrary:Register(Spell, MAJOR_VERSION, MINOR_VERSION, activate, deactivate, external)
Spell = nil